/*******************************************************************************
 * Copyright (c) 2010- Association for Decentralized Information Management in
 * Industry THTH ry.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *    VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard;

import java.io.File;
import java.io.IOException;

import org.simantics.databoard.accessor.RecordAccessor;
import org.simantics.databoard.accessor.error.AccessorConstructionException;
import org.simantics.databoard.accessor.error.AccessorException;
import org.simantics.databoard.accessor.file.FileVariantAccessor;
import org.simantics.databoard.accessor.impl.ClosableCompositeRecord;
import org.simantics.databoard.accessor.impl.DirectoryMap;
import org.simantics.databoard.accessor.impl.ClosableCompositeRecord.CloseListener;
import org.simantics.databoard.accessor.impl.CompositeRecord;
import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.binding.error.RuntimeBindingException;
import org.simantics.databoard.serialization.RuntimeSerializationException;
import org.simantics.databoard.serialization.SerializationException;
import org.simantics.databoard.type.ArrayType;
import org.simantics.databoard.type.DataType;
import org.simantics.databoard.type.IntegerType;
import org.simantics.databoard.type.MapType;
import org.simantics.databoard.type.RecordType;
import org.simantics.databoard.type.UnionType;
import org.simantics.databoard.type.VariantType;

/**
 * Fasade Class for Managing historian data 
 *
 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
 */
public class Historian {
	
	public static DataType getRecordingSessionType() {
		return DataTypes.getDataType("RecordingSession");
	}
	
	/**
	 * Create a new directory based historian. 
	 * There is a session.dbb file that contains RecordingSession info.
	 * The recordings field is empty as they are separate .dbb files.<p>
	 * 
	 * IOException is thrown if directory already existed, otherwise the
	 * directory is used. <p>
	 * 
	 * @param directory
	 * @throws IOException
	 */
	public static void createDirectoryHistorian(File directory) 
	throws IOException {
		if (directory.exists())
			throw new IOException(directory+" already exists");
		try {
			DataType recordingSession = DataTypes.getDataType( "RecordingSession" );
			File file = new File(directory, "session.dbb");
			Files.createFile(file, recordingSession);
		} catch (BindingException e) {
			throw new RuntimeBindingException(e);
		} catch (SerializationException e) {
			throw new RuntimeSerializationException(e);
		}
	}
	
	/**
	 * Open an accessor to directory & file based historian service.
	 * Note atmost one accessor is allowed. 
	 * 
	 * @param directory 
	 * @return accessor
	 * @throws AccessorConstructionException 
	 */
	public static RecordAccessor openDirectoryHistorian(File directory) throws AccessorConstructionException {
		// Open file
		File file = new File(directory, "session.dbb");
		final FileVariantAccessor va = Accessors.openAccessor(file);
		RecordAccessor ra = va.getContentAccessor();

		// Remap recordings to directory
		final DirectoryMap dm = new DirectoryMap(directory);

		// Forward close to file & Directory 
		CloseListener closeListener = new CloseListener() {
			public void onClosed(CompositeRecord sender) throws AccessorException {
				try {
					va.close();
				} finally {
					dm.close();
				}
			}
		};
		
		// Remap session.dbb and the directory to composite record
		CompositeRecord cr = new ClosableCompositeRecord(closeListener);
		cr.addField("datasource", ra.getFieldAccessor("datasource"));
		cr.addField("events", ra.getFieldAccessor("events"));
		cr.addField("milestones", ra.getFieldAccessor("milestones"));
		cr.addField("timeVariableId", ra.getFieldAccessor("timeVariableId"));
		cr.addField("recordings", dm);		
						
		return cr;
	}

	/**
	 * Create DataType for Recording(T, V) where T is the datatype of the Time, and V value.
	 * 
	 *  // The captured data for a single variable
     *  type Recording(T, V) = {
     *         // Record Identifier
     *         nodeId : Variant,
     *        
     *         // All labels
     *         labels : LocalizedTexts,
     * 
     *         // All Segments, segments are time series
     *         segments : Map(T, V)[]
     *       }
	 * 
	 * @param T value type
	 * @return Recording(T)
	 */
	public static DataType createRecordingType(DataType timeType, DataType valueType) {
		RecordType rt = new RecordType();

		// Recording(T)
		rt.addComponent("nodeId", VariantType.INSTANCE);
		rt.addComponent("labels", VariantType.INSTANCE);
		rt.addComponent("segments", new ArrayType( new MapType( timeType, valueType ) ) );		
		
		return rt;
	}

	
}

