/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.reflection;

import java.util.Map;

import org.simantics.databoard.annotations.Union;
import org.simantics.databoard.binding.Binding;
import org.simantics.databoard.binding.UnionBinding;
import org.simantics.databoard.binding.error.BindingConstructionException;
import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.reflection.ReflectionBinding.ClassBinding;
import org.simantics.databoard.reflection.ReflectionBinding.ClassDataType;
import org.simantics.databoard.type.DataType;
import org.simantics.databoard.type.UnionType;

/**
 * Binds a inheritance of classes into a union of records. 
 *
 *
 */
public class ReflectionUnionClassBinding extends UnionBinding {

	Class<?> clazz;
	Class<?>[] classes;
	
	/**
	 * Create a UnionType to Abstract Class Binding. 
	 * 
	 * The construction is 2-phase. 
	 * The second phase (#postConstruction) adds sub class bindings.
	 * 
	 * @param type
	 * @param clazz
	 * @throws BindingConstructionException
	 */
	public ReflectionUnionClassBinding(UnionType type, Class<?> clazz) 
	throws BindingConstructionException {
		this.type = type;
		this.clazz = clazz;
		Union unionAnnotation = clazz.getAnnotation(Union.class);
		if (unionAnnotation==null) 
			throw new BindingConstructionException("Union annotation is required in "+clazz.getCanonicalName());
		
    	classes = unionAnnotation.value();
    	componentBindings = new Binding[ classes.length ];
	}
	
	/**
	 * The second phase of the 2-phase construction.
	 * @throws BindingConstructionException 
	 */
	public void postConstruction(ReflectionBinding bindingScheme, Map<ClassBinding, Binding> map) throws BindingConstructionException {
		UnionType ut = type();
       	for (int i=0; i<count(); i++) {
       		DataType dt = ut.components[i].type;
       		Class<?> c = classes[i];
    		ClassBinding componentBindingRequest = new ClassBinding( dt, new ClassDataType(c) );
    		componentBindings[i] = bindingScheme.unsynchronizedGetBinding(componentBindingRequest, map);				
       	}		
	}
	
	@Override
	public Object create(int tag, Object value) {
		return value;
	}

    @Override
    public void setValue(Object union, int tag, Object value)
	throws BindingException {
    	throw new BindingException("Cannot change the class of an instance");
    }
	
	@Override
	public int getTag(Object obj) throws BindingException {
		for (int i=0; i<classes.length; i++)
			if (classes[i].isInstance(obj)) return i;
		throw new BindingException(obj+" is not an element of "+clazz.getSimpleName());
	}
	@Override
	public Object getValue(Object obj) {
		return obj;
	}
	@Override
	public boolean isInstance(Object obj) {
		for (Class<?> c : classes) 
			if (c.isInstance(obj)) return true;					
		return false;
	}
	
	@Override
	public boolean isImmutable() {
		return true;
	}
	
}